//
// $Id: Empire.m,v 1.7 1997/10/31 04:51:47 nygard Exp $
//

//
//  This file is a part of Empire, a game of exploration and conquest.
//  Copyright (C) 1996  Steve Nygard
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  You may contact the author by:
//     e-mail:  nygard@telusplanet.net
//

#import "Empire.h"

RCSID ("$Id: Empire.m,v 1.7 1997/10/31 04:51:47 nygard Exp $");

#import <AppKit/AppKit.h>

//----------------------------------------------------------------------
// Functions for manipulating MapTokens and their components
//----------------------------------------------------------------------

MapToken EMCreateMapToken (Terrain terrain, Player player, Icon icon, BOOL explored)
{
    return ((explored == YES) ? 0 : MT_EXPLORED_MASK)
        | ((player << 6) & MT_PLAYER_MASK)
        | ((icon << 2) & MT_ICON_MASK)
        | (terrain & MT_TERRAIN_MASK);
}

//----------------------------------------------------------------------

void EMMapTokenComponents (MapToken mapToken, Terrain *terrain, Player *player, Icon *icon, BOOL *explored)
{
    if (terrain != NULL)
        *terrain = mapToken & MT_TERRAIN_MASK;

    if (player != NULL)
        *player = (mapToken & MT_PLAYER_MASK) >> 6;

    if (icon != NULL)
        *icon = (mapToken & MT_ICON_MASK) >> 2 ;

    if (explored != NULL)
        *explored = ((mapToken & MT_EXPLORED_MASK) == 0) ? YES : NO;
}

//----------------------------------------------------------------------

Terrain EMTerrainComponent (MapToken mapToken)
{
    return mapToken & MT_TERRAIN_MASK;
}

//----------------------------------------------------------------------

Player EMPlayerComponent (MapToken mapToken)
{
    return (mapToken & MT_PLAYER_MASK) >> 6;
}

//----------------------------------------------------------------------

Icon EMIconComponent (MapToken mapToken)
{
    return (mapToken & MT_ICON_MASK) >> 2;
}

//----------------------------------------------------------------------

BOOL EMExploredComponent (MapToken mapToken)
{
    return ((mapToken & MT_EXPLORED_MASK) == 0) ? YES : NO;
}

//----------------------------------------------------------------------

MapToken EMChangeTerrainComponent (MapToken mapToken, Terrain terrain)
{
    return (mapToken & ~MT_TERRAIN_MASK) | (terrain & MT_TERRAIN_MASK);
}

//----------------------------------------------------------------------

MapToken EMChangePlayerComponent (MapToken mapToken, Player player)
{
    return (mapToken & ~MT_PLAYER_MASK) | ((player << 6) & MT_PLAYER_MASK);
}

//----------------------------------------------------------------------

MapToken EMChangeIconComponent (MapToken mapToken, Icon icon)
{
    return (mapToken & ~MT_ICON_MASK) | ((icon << 2) & MT_ICON_MASK);
}

//----------------------------------------------------------------------

MapToken EMChangeExploredComponent (MapToken mapToken, BOOL explored)
{
    return (mapToken & ~MT_EXPLORED_MASK) | ((explored == YES) ? 0 : MT_EXPLORED_MASK);
}

//----------------------------------------------------------------------

UnitType EMConvertIconToUnitType (Icon icon)
{
    static const unitTypesForIcons[] = {u_unknown, u_army, u_army, u_fighter, u_transport, u_transport, u_submarine,
                                        u_destroyer, u_cruiser, u_carrier, u_carrier, u_battleship, u_hovercraft};

    NSCAssert (icon <= i_hovercraft, @"Icon out of range.");

    return unitTypesForIcons[icon];
}

//----------------------------------------------------------------------

Icon EMConvertUnitTypeToBaseIcon (UnitType unitType)
{
    static const Icon baseIconsForUnitTypes[] = {i_none, i_army, i_fighter, i_unloaded_transport, i_submarine,
                                                 i_destroyer, i_cruiser, i_unloaded_carrier, i_battleship, i_hovercraft};

    NSCAssert (unitType <= u_hovercraft, @"UnitType out of range.");

    return baseIconsForUnitTypes[unitType];
}

//----------------------------------------------------------------------
// Other utility functions for Empire types
//----------------------------------------------------------------------

NSString *EMUnitTypeName (UnitType aType)
{
    NSString *unitTypeNames[] = {@"Unknown", @"Army", @"Fighter", @"Transport", @"Submarine", @"Destroyer", @"Cruiser",
                                 @"Carrier", @"Battleship", @"Hovercraft"};

    NSCAssert (aType <= u_hovercraft, @"UnitType out of range.");

    return unitTypeNames[aType];
}

//----------------------------------------------------------------------

NSString *EMShortDirectionName (Direction dir)
{
    NSString *short_names[] = {
        @"NW", @"N", @"NE", @"W", @"None", @"E", @"SW", @"S", @"SE"
    };

    NSCAssert (dir <= d_southeast, @"Direction out of range." );

    return short_names[dir];
}

//----------------------------------------------------------------------

NSString *EMLongDirectionName (Direction dir)
{
    NSString *long_names[] = {
        @"Northwest", @"North", @"Northeast", @"West", @"None", @"East", @"Southwest", @"South", @"Southeast"
    };

    NSCAssert (dir <= d_southeast, @"Direction out of range." );

    return long_names[dir];
}

//----------------------------------------------------------------------

BOOL EMUnitTypeCanSee (UnitType viewer, UnitType viewed)
{
    // Subs cannot see army or fighter
    // only sub, destroyer, or cruiser can see sub

    if ((viewer == u_submarine && (viewed == u_army || viewed == u_fighter))
        || (viewed == u_submarine && (viewer != u_submarine
                                      && viewer != u_destroyer
                                      && viewer != u_cruiser)))
    {
        return NO;
    }

    return YES;
}

//----------------------------------------------------------------------
// Calculate the distance (ignoring terrain) between two locations.
// This is just the greatest differene between rows and columns.
//----------------------------------------------------------------------

int EMDistanceBetweenLocations (EMMapLocation location1, EMMapLocation location2)
{
    int dr, dc;
    int distance;

    dr = location1.row - location2.row;
    dc = location1.column - location2.column;

    dr = (dr < 0) ? -dr : dr;
    dc = (dc < 0) ? -dc : dc;
        
    distance = (dr > dc) ? dr : dc;

    return distance;
}

//----------------------------------------------------------------------
// Functions to assist in debugging
//----------------------------------------------------------------------

NSString *EMFormatComponents (MapToken mapToken)
{
    Terrain terrain;
    Player player;
    Icon icon;
    BOOL explored;
    UnitType unitType;
    char *unitReps = "-aftsdcCbh";
    char *terrainReps = " wl#";

    EMMapTokenComponents (mapToken, &terrain, &player, &icon, &explored);
    unitType = EMConvertIconToUnitType (icon);

    return [NSString stringWithFormat:@"%@%x%c%c",
                     explored ? @"Y" : @"N",
                     player,
                     unitReps [unitType],
                     terrainReps [terrain]];
}

//----------------------------------------------------------------------

NSString *EMFormatNineComponents (MapToken *tokens)
{
    return [NSString stringWithFormat:@"\n%@ %@ %@\n%@ %@ %@\n%@ %@ %@",
                     EMFormatComponents (tokens[0]),
                     EMFormatComponents (tokens[1]),
                     EMFormatComponents (tokens[2]),
                     EMFormatComponents (tokens[3]),
                     EMFormatComponents (tokens[4]),
                     EMFormatComponents (tokens[5]),
                     EMFormatComponents (tokens[6]),
                     EMFormatComponents (tokens[7]),
                     EMFormatComponents (tokens[8])];
}

//----------------------------------------------------------------------

NSString *EMFormat25Components (MapToken *tokens)
{
    return [NSString stringWithFormat:@"\n%@ %@ %@ %@ %@\n%@ %@ %@ %@ %@\n%@ %@ %@ %@ %@\n%@ %@ %@ %@ %@\n%@ %@ %@ %@ %@",
                     EMFormatComponents (tokens[0]),
                     EMFormatComponents (tokens[1]),
                     EMFormatComponents (tokens[2]),
                     EMFormatComponents (tokens[3]),
                     EMFormatComponents (tokens[4]),
                     EMFormatComponents (tokens[5]),
                     EMFormatComponents (tokens[6]),
                     EMFormatComponents (tokens[7]),
                     EMFormatComponents (tokens[8]),
                     EMFormatComponents (tokens[9]),
                     EMFormatComponents (tokens[10]),
                     EMFormatComponents (tokens[11]),
                     EMFormatComponents (tokens[12]),
                     EMFormatComponents (tokens[13]),
                     EMFormatComponents (tokens[14]),
                     EMFormatComponents (tokens[15]),
                     EMFormatComponents (tokens[16]),
                     EMFormatComponents (tokens[17]),
                     EMFormatComponents (tokens[18]),
                     EMFormatComponents (tokens[19]),
                     EMFormatComponents (tokens[20]),
                     EMFormatComponents (tokens[21]),
                     EMFormatComponents (tokens[22]),
                     EMFormatComponents (tokens[23]),
                     EMFormatComponents (tokens[24])];
}

//----------------------------------------------------------------------

NSString *EMFormatVisibilityFlags (int flags)
{
    return [NSString stringWithFormat:@"%@%@%@%@",
                     (flags & VIS_SUB)  ? @"U" : @"-",
                     (flags & VIS_LAND) ? @"L" : @"-",
                     (flags & VIS_SEA)  ? @"S" : @"-",
                     (flags & VIS_AIR)  ? @"A" : @"-"];
}

//----------------------------------------------------------------------

NSString *EMFormatNineVisibilityFlags (int *flags)
{
    return [NSString stringWithFormat:@"\n%@ %@ %@\n%@ %@ %@\n%@ %@ %@",
                     EMFormatVisibilityFlags (flags[0]),
                     EMFormatVisibilityFlags (flags[1]),
                     EMFormatVisibilityFlags (flags[2]),
                     EMFormatVisibilityFlags (flags[3]),
                     EMFormatVisibilityFlags (flags[4]),
                     EMFormatVisibilityFlags (flags[5]),
                     EMFormatVisibilityFlags (flags[6]),
                     EMFormatVisibilityFlags (flags[7]),
                     EMFormatVisibilityFlags (flags[8])];
}

//----------------------------------------------------------------------

NSString *EMFormat25VisibilityFlags (int *flags)
{
    return [NSString stringWithFormat:@"\n%@ %@ %@ %@ %@\n%@ %@ %@ %@ %@\n%@ %@ %@ %@ %@\n%@ %@ %@ %@ %@\n%@ %@ %@ %@ %@",
                     EMFormatVisibilityFlags (flags[0]),
                     EMFormatVisibilityFlags (flags[1]),
                     EMFormatVisibilityFlags (flags[2]),
                     EMFormatVisibilityFlags (flags[3]),
                     EMFormatVisibilityFlags (flags[4]),
                     EMFormatVisibilityFlags (flags[5]),
                     EMFormatVisibilityFlags (flags[6]),
                     EMFormatVisibilityFlags (flags[7]),
                     EMFormatVisibilityFlags (flags[8]),
                     EMFormatVisibilityFlags (flags[9]),
                     EMFormatVisibilityFlags (flags[10]),
                     EMFormatVisibilityFlags (flags[11]),
                     EMFormatVisibilityFlags (flags[12]),
                     EMFormatVisibilityFlags (flags[13]),
                     EMFormatVisibilityFlags (flags[14]),
                     EMFormatVisibilityFlags (flags[15]),
                     EMFormatVisibilityFlags (flags[16]),
                     EMFormatVisibilityFlags (flags[17]),
                     EMFormatVisibilityFlags (flags[18]),
                     EMFormatVisibilityFlags (flags[19]),
                     EMFormatVisibilityFlags (flags[20]),
                     EMFormatVisibilityFlags (flags[21]),
                     EMFormatVisibilityFlags (flags[22]),
                     EMFormatVisibilityFlags (flags[23]),
                     EMFormatVisibilityFlags (flags[24])];
}

//======================================================================
// Notifications
//======================================================================

NSString *EMPlayerHasWonGameNotification   = @"EMPlayerHasWonGameNotification";
NSString *EMGameStateChangedNotification   = @"EMGameStateChangedNotification";
NSString *EMTurnChangedNotification        = @"EMTurnChangedNotification";
NSString *EMPlayerStateChangedNotification = @"EMPlayerStateChangedNotification";
